'''
This module is used to automatically update the data in a standalone API instance.
 (e.g. used in biothings.standalone, now hub.standalone module, and biothings.farms repo)
'''

from .dumper import BiothingsDumper  # noqa: F401
from .uploader import BiothingsUploader  # noqa: F401
